'*********************************************************
'*                                                       *
'*   ZiCog_cpm v1.0_rc_5.5                               *
'*                                                       *
'*   A simulation of a Zilog Z80 based computer system   *
'*   Capable of running CP/M (v2.2 and 3)                *
'*                                                       *
'*                                                       *
'*   (C) 2009 Michael Rychlik, Ray Roderick and J Moxham *
'*                                                       *
'*********************************************************
'  
'v0.0  -  1 Mar 2009 - Original version
'
'v0.1  -  4 Mar 2009 -
'
'v0.2  -  7 Mar 2009 -
'
'v0.3  -  7 Mar 2009 -
'
'v0.4  - 16 Mar 2009 - Replaced T8080 diagnostic with ZICOGTST.COM. This is a version of EXZ80DOC.COM from
'                      the SIMH AltairZ80 simulator with undocumented instructions removed and tests rearranged
'                      in execution time order (almost). T8080.COM no longer passes tests because due to the
'                      implementation of Z80 flag settings rather than 8080.
'                      Fixed up SIMH I/O simulation for console, disks on SD card etc.
'
'v0.5  - 18 Mar 2009 -
'
'v0.6  - 30 Mar 2009 - Removed "outa" capability in output emulation.
'
'v0.7  - 13 Mar 2009 - Set to use zicog version 0.7.
'
'v0.8  -  3 May 2009 - Added SIMH AltairZ80 style hard disk emulation.
'                      Conformant with SIMH ALTAIR 8800, BIOS V1.25, 1 HD, 19-Apr-2009
'
'v0.9  -  5 May 2009 - Added conditional compilation to select Z80 or 8080 emulation. (#ifdef CPU_Z80/CPU_8080)
'                    - Added PropDemoBoard as a conditional compilation option and checks for missing definitions.
'                    - Moved configuration #defines to zicog_demo so they are all in one place.
'                    - Fixed hard disk reset command.
'                    - Added DPB definition for 720K Super I/O Floppy, for use as RAM disk.
'                    - Added some sanity checking on selected HDSK drive, track and sectors.
'                    - Added conditional compilation (#ifdef HEARTBEAT) to enable hearbeat LED flashing from Z80 emulator
'
'v0.10 - 19 Aug 2009 - Incorporate all Cluso's and DrAcula's changes
'                      from zicog_demo009_rr097.spin as listed below.
'                    - Removed keyboard object and all references to kb.something
'                    - Removed TV object and all references to TV.something
'                    - Replaced kb and TV with UART using FullDuplexSerial
'                    - Removed FrequencyCount object and all reference to fcounter (no one was using it)
'                    - Removed a "repeat" from in_hdskport after "HDSK ERROR 5", saves the survey program hanging.
'                    - Changed in_condata to use UART.rxcheck, saves the survey program from hanging.
'                    - Added disk I: to the disk start up message.
'                    - It is now necessary to use SPACE to single step instructions.
'
'1.0_rc_2 - 28 Aug 2009 - Changed to use 4 sectors of 128 bytes per SD block instead of just 1.
'                    - 1 SD block is held in a "cache" at any time so that reads from sectors
'                    - already in the cache do not need to access the SD. Writes to sectors currently in
'                    - the cache do not need to be preceded by sector fetches.
'
'1.0_rc_3 - 5 Sep 2009 - Moved definition of the Z80 RAM area (ram_z80) from zicog_demo.spin
'                      to zicog.spin and placed the Z80 emulation PASM code inside it!.
'                      This save ~2K HUB RAM as the space occupied by the emulator COG code
'                      is now recycled as Z80 RAM when the COG is up and running.
'                      This results in being able to run 24K CP/M system on a Prop Demo board again
'                      which in turn allows booting from disk again.
'                    - Added disk parameters for 440K RAM disk as dpb_3.     <==================???? dpb_6 ???
'                    - Removed all code for handling Altair 1M floppy disks.
'                      Brutal I know but they were messy to use with SD cards and very slow.
'                      If there are anyone really wants them back we can dig the code out of the archive
'                    - File renamed zicog_cpm.spin (from zicog_demo.spin)
'                    - Removed old cpu_control cpu parameter
'                    - Removed ram and rom junk from cpu parameters
'1.0_rc_4 - 7 Sep 2009 -
'
'1.0_rc_5 - 8 Sep 2009 - Made at least compilable for TriBlade and CPU_Z80
'                      - Changed sd_block to sd_block_numer throughout TriBlade code.
'                      - Added warning if no platform define is specified.
'
'1.0_rc_5.5-8 Sep 2009 - Added LoadZ80Ram to read disk boot sectors to Z80 RAM
'                        prior to start.
'                        Requires new BOOTGEN to lay down those sectors.
'
'' RR20090404   Fix required sections for compiling by homespun024.exe
''              sdspiqasm, PC_Keyboard, PC_Text, FrequencyCount all reguired org -> org 0
''              zicog: jmp requires # for line 374, and a description for jmp without # to say jmp/ret(indirect) 
''
'' +-------------------------------------------------------------------------------------------------------+
'' | Modifications to ZiCog v0.5 for use on Cluso's TriBladeProp PCB (Blade #2)                       vxxx |
'' +-------------------------------------------------------------------------------------------------------+
'' |  Author:            "Cluso99" (Ray Rodrick)                                                           |
'' |  Copyright (c) 2009 "Cluso99" (Ray Rodrick)                                                           |
'' |  License            Modifications only under MIT License                                              |
'' +-------------------------------------------------------------------------------------------------------+
'' RR20090317   Cluso99: Modifications to ZiCog v0.5 for use on the TriBladeProp PCB (Blade #2)
'' RR20090317   start to use TriBladeProp & SRAM                             
''              Ram is only $4F00 - shouldn't it be $5000 ??                 
''              Remove heartbeat and FrequencyCounter.spin and fcounter.xxx calls
''              Remove sdspiqasm.spin and disk.xxx calls                     
''              Remove output LEDs                                           
'' RR20090318   add sram_start (ram offset)
''              Rom is $0180 long beginning $FF00 - shouldn't it be $0100 long ??
'' RR20090318   move from v0.4 to v0.5
'' RR20090318   _rr004 version
'' RR20090319   make 64KB CPM version
''              change msize 20 -> 64; disable free_ reservation in hub
'' RR20090320   _rr006 version: add SRAM disk
''              _rr007 complete and enable SRAM disk
'' RR20090321   _rr008 use new CBIOSX_triblade.COM
''              preload SRAM disk from cpm2.dsk using cpm2SD_sram_rr008 (truncates)
''              enabled readblock & writeblock (no clear on start). DIR -> "no file"
'' RR20090321   _rr009 incorporate cpm2SD_sram_rr008 to preload SRAM with cpm disk
''              short of stack space - disable part of print_regs debug code
''              _rr010 DIR says "no file"
''              _rr011 use sdsram_rr011 for loading xxxx.COM files to Z80RAM (64KB)
''              _rr015 preload SRAM disk first using cpm2SD_debug4 & SRAM_driver_015
'' RR20090331   _rr021 use TriBladeProp_Blade2_201.spin driver to load Z80 Ram and CPM Ram Disk
''              _rr022 bugfix (sector 127->128) Use Cluso's VB6 to load Ram from PC
''              _rr023 works loading SRAM and verified (using VB) for Z80 & Ram Disk 448_512KB
''                     gives A> prompt but DIR still does not work
''              _rr024 LoadSRAM optional constant
''              _rr025 the Ram Disk files have 137 byte sectors (not 128 as first thought) so fix in spin for now
''                     Hooray it does DIR :-)
''                     Warning, likely to require more stack space before continuing !!!
'' RR20090401   _rr026 start to load from microSD the SRAM 1MB "DRIVE_1M.RAM" (uses first 128 bytes of 512 byte sectors)
''              _rr027 remove the cpm code included files (z80ram tpa, ccp, etc) to freeup space
'' RR20090402   _rr028 add SD_Trainer_rr008 "FullDuplexSerialPlus" & "fsrwFemto" for microSD load to SRAM
''              _rr034 writing issue - try new driver _v202 (not the problem)
'' RR20090403   _rr035 (proved sram walking code ok, so some interaction with sd card drivers?)
''              _rr036 Fixed bug: SD card not tristating D0 - fixed sdspiFemto_rr001.spin
''              _rr040 Loading 1MB SRAM from "DRIVE_1M.RAM" on microSD and getting A>
''              _rr041 Does a DIR listing (SRAM loaded, so loading commented out)
'' RR20090404   _rr043 Migrate to ZiCog v0.006 & homespun compiler
'' ---------------------------------------------------------------------------------------------------------
'' RR20090404   _rr044 Migrate to ZiCog v0.006 & homespun compiler
''              _rr045 question to load SRAM from microSD, Fix ram_size
''              _rr045 Write working on _rr042(later), added here & works
''                     Note does not work with zicog_rr044 but does with zicog005_rr021
'' RR20090405   _rr046 Remove clear screen for loading from PropCMD - working & posted
''              _rr047 remove special prop UTF-16 characters
''              _rr048 done
'' RR20090407   _rr049 modify to use direct sector address for disk files on microSD card
''              _rr050 loading the 64KB+960KB using direct SD sector addressing
'' RR20090408   _rr051 convert keys
''              _rr052 slow to 9600 baud and do comms between Blades #2<-->#1
''              _rr053         4800
'' ---------------------------------------------------------------------------------------------------------
'' SKIP _rr052 & _rr053 (for now)
'' RR20090416    _rr055 (from _rr051) clear sram 0-64KB and only load $FF00-$FFFF DSKBOOT.COM
''                      now CPM should boot from the sram disk - does not work, so place jmp to $FF00 at $0000
'' RR20090416    _rr056 working (removed jmp $FF00 at $0000)
''               _rr057 get 2nd disk running (on SD card)
'' RR20090417    _rr058 continue
''               _rr059 continue
''               _rr060 continue (new SD files DSK_CPM2.DSK & DSK_CPM3.DSK) older formats incorrect
''                      loads ok but will not take kbd input due to restarts???
''               _rr061 continue (add #define sddisk)
''               _rr062 conflict (add #define sduseinit) prevent readsd from rebooting cog each time - works
''               _rr063 conflict (#define sddisk) - no kbd input at A> - now works, but resetting tbp2 cog !!
''                      but not using the sd disk data yet
''               _rr064 use external CPM2 disk on sd card (#define sdCPM2)
''               _rr065 (#define RamDisk) Note to fix resetting RamLatches from resetting when accessing sdcard
''                      working
''               _rr066 tidy code (RamLatches driver requires "idle" command instead of start/stop)
''                      add drive1base and 2 disks   - works
'' RR20090420    _rr067 try to do syscpm2 - works
'' RR20090421    _rr068 add drives C...H
'' RR20090421    _rr069 try cpm3
''                      try heaters new nonbanked cpm3 - works but have to enter 2 characters to get 1 input !!
'' RR20090429    _rr070 fixed RamLatches idle/active for bus conflict "TriBladeProp_Blade2_202"
''               _rr072 add #define CPM3 (not tested)
'' RR20090430    _rr073 new nonbanksleep CMP3 version  (doesn't fix keyb bug)
''                      try and use ZiCog v0.07_rr070
'' RR20090504    _rr074 try ZiCog v0.08_rr073 (_rr080 working)
'' ---------------------------------------------------------------------------------------------------------
'' RR20090504    _rr080 combine v0.08 demo with demo_rr074
''                      CPM3 still has kbd bug
''               _rr081 add #define hdskenable
'' ---------------------------------------------------------------------------------------------------------
'' RR20090507    _rr090 combine v0.09
'' RR20090805    _rr091 use standalone in Blade #2 (released 20090805)
'' RR20090806    _rr092 use PC_Text & PC_Keyboard (replace FDXplus) 
''               _rr093 add #define addLF
'' RR20090817    _rr094 Dr_Acula: N8VEM & fix dropping received characters (_rr092_james2)
'' RR20090818    _rr095 fix reversed rxavail
''               _rr096 add 4 x Hard Disk support (I:,J:,K:,L:)  partial only
'' Dr20090818    _rr097 Dr_Acula: modifications for CRLF (_rr096_j18)
'' ---------------------------------------------------------------------------------------------------------
'' MR20090819    v0.10  modifications by heater (see above)
'' ---------------------------------------------------------------------------------------------------------
'' RR20090819    _rr100 add #define fix137 (for the fudge 137 byte sectors for SIMH); minor code tidyup
'' RR20090819    _rr101 remove fix137/preset_137 not required ($A7,$0E,$D2.....$B7,$C8,$CD,$A5,$2B,$00)
'' RR20090820    _rr102 many thanks to Dr_Acula and Heater for the combined testing evening.
''                      fix hard disk on tbp
''                      force jmp to $FF00 (faster)
''                      some tidyup
'' RR20090821    _rr103 allow 26 drive(SD file offset) parameters
''                      R: is RamDisk
'' RR20090823    _rr104 Note 16 drives is the max for CPM, SIMH has 12 (A:..H: + I:..L:)
''                       so allow 8x1MB Floppy + 4x HardDisk (2x8MB, 1x..KB Ram, 1x..MB large)
'' RR20090829    _rr105 Use "DSKBOOT_.ROM" to boot (load into $FF00)
''                      Use HDRIVE_I.DSK & HDRIVE_J.DSK filled with "$E5" and then PIP I:=A:*.*
''               _rr106 Setup RamDisk F: with "$E5" - works
''               _rr107 Make RamDisk hard disk J: - works BUT bounds checking invalid (see STAT)
''                      General code tidy, change to common CheckError routine and check <0
''                      Note: hdp block size mods do not work
'' ---------------------------------------------------------------------------------------------------------
'' MR20090907  v1.0_rc4 modifications by heater (see above)
''                      rename ZiCog_demo -> ZiCog_cpm
'' ---------------------------------------------------------------------------------------------------------
'' RR20090907    _rr108 add heaters mods (fails compilation)
'' RR20090908    _rr109 try make it compile    
'' ---------------------------------------------------------------------------------------------------------
'' 1.0_rc_5.5b-8 Sep 2009  modifications by heater (see above)
'' RR20090913    _rr112 mod to pass all SD pins to sdspiqasm_rr001
'' RR20090913    _rr113 now runs 25K CPM2.2 version on TriBlade hardware WITHOUT USING SRAM.
'' RR20090913    _rr114 TriBlade to use SRAM again (blocking & zipm2.dsk)
'' RR20090914    _rr116 TriBlade runs 25K CPM2.2 version with sram
'' RR20090914    _rr117 PropDemoBoard - code disk offsets into the file using
'' RR20090915    _rr118 TriBlade & 64K CPM2.2 version with sram (reqs Z80 emulation)
''               _rr119 Try CPM2___I.DSK & CPM3___I.DSK
''               _rr120 Add back floppy support for TriBlade only (to PIP I:=B:*.* for converting floppy to hdisk)
''               _rr121
'' RR20090917    _rr122 Use new fsrw23 routines (fsrw.spin & mb_rawb_spi.spin)
''                        modified to disable readahead and release SD pins after use
''                        fsrwFemto_rr001 for compatibility with new fsrw
''                        bug in large copies >23KB (problem in 64KB build of CPM disk)
'' RR20090922    _rr123 96MHz (6MHz xtal)
''                        remove block_offset from refresh_hd_cache
''
'' RR20091003    _rr125 Use new fsrw23 routines
''               _rr126
''
''
''
''      complete rewrite late November 2009 Dr_Acula (James Moxham) for running on the single propeller DracBlade
'' includes VT100 driver, vga screen, keyboard, 20x4 LCD display, sd card
' 5th Jan 2010 added port drivers for port 2
' tweaked the vga driver so the " character displays better
' note needs BST to compile and needs Tools/Compiler Preferences/Optimisations/Eliminate unused spin methods, otherwise runs out of memory
' 7th Jan added baud rate change code for port 2 so can change from cp/m
' fixed a bug where caps lock was printing ^, this is now captured in GotKey of the keyboard object
''
''      Modified 28 Mar 2009 hairymnstr (Nathan Dumont)
''      Moved DracBlade RAM access functions into the SD cog see sdspiDrac.spin
''
CON

  baud           = 19200 ' max speed for xmodem with the buffer size of the 4 port serial object (? could increase the buffer size)
'  baud          = 38400

'   baud         = 1200  ' for very slow testing

'Uncomment one of the two following CPU emulation types.
#define CPU_Z80
#define DracbladeProp ' dracblade options in zicog
'#define CPU_8080

'#define SingleStep                                     'Uncomment if "single step" emulation required.
'#define banked_memory

  _clkmode      = xtal1 + pll16x
  _xinfreq      = 5_000_000

  spiDO         = 12                                     ' microSD pins
  spiClk        = 13
  spiDI         = 14
  spiCS         = 15

  cpmsize       = 64   * 1024                           'CPM Ram size (see msize below)
  sramsize      = 512  * 1024                           '512KB SRAM

' VGA and keyboard for DracBoard
  Cursor     = 95
  VideoCls   = 0
  NUM        = %100
  CAPS       = %010
  SCROLL     = %001
  RepeatRate = 40
  video      = 16

'Z80 Program Status Word bits
sign_bit        = %10000000
zero_bit        = %01000000
aux_bit         = %00010000
parity_bit      = %00000100
carry_bit       = %00000001


'I/O control block commands
io_cmd_out      = $01
io_cmd_in       = $02
io_cmd_break    = $03

'
'       A read to the data port gets the buffered character, a write
'       to the data port writes the character to the device.
constat         = $10                   ' (16)sio port 1 status port - a combo of keyboard and both serial ports
condata         = $11                    ' (17)sio port 1 data port - a combo of keyboard and both serial ports
punstat         = $12                    ' (18)sio port 2 status port
pundata         = $13                    ' (19)sio port 2 data port

simhport        = $FE
srport          = $FF                  ' Altair front panel switch register port

ledport         = $00



' sd card file access port
myname_port     = $70                   ' hopefully not used by anything else
port2_Off       = $71                   ' a little CP/M program sends an OUT to port &H71
Forecolor       = $72                   ' set foreground color
Backcolor       = $73                   ' set background color
keystat         = $74                   ' test just keyboard status (see also constat above which does this too as a combo)
keydata         = $75                   ' get just the keyboard character
'ram             = $76                   ' for I/O to 512k ram - works but commented out as not used much

' UART constants for N8VEM
n8vem_uart_test = $6D
n8vem_uart_in   = $68
n8vem_uart_out  = $68

' UART constants for the second port (male D9)
port2_test      = $7D                    ' to test if a byte is ready
port2_io        = $78                    ' for input and output to port 2
port2_baud      = $7C                    ' set baud rate multiple of 1200 eg 1=1200, 16=19200, 96=115200

' masks for SIO controller (constat, punstat)
mout            = $02                   ' output allowed mask

' commands for SIO controller (constat, punstat)
creset          = 3                     ' reset command

' constants for hard disk port
nhdisks         = 8                     'The number of supported hard disks

hdskNull        = 0                     ' No command
hdskReset       = 1                     ' command to reset controller
hdskRead        = 2                     ' read command
hdskWrite       = 3                     ' write command
hdskParam       = 4                     ' Get HD parameters command
hdskport        = $fd                   ' control port for simulated hard disk

hdsk_base_block = 0                     ' First block of first hard disk in SD.

'constants for SIMH port
simhreset       = 14
stoptimer       = 22
setbankselect   = 12
hasbankedmemory = 18

VAR
  Word Key
  Byte Keyflag                    ' flag to say a keyboard character is available
  Byte UART0flag                  ' flag to say a character is available on serial port 0
  Byte UART1flag                  ' flag to say a character is available on serial port 1 (port2)
  Byte termcolor
  Byte curset
  Byte TerminalFlag               ' Running as a terminal for other boards -1=true, 0=false
  Long HighLatchByte              ' xxxxxxxx_00000000_xxxxxxxx_xxxxxxxx ' A16-A19 and led and spare 4 outputs
  byte Myname[9]                   ' store my name in a tiny array
  long OldCnt1                     ' sd card countdown
  long Countdown1                  ' sd card countdown
  long OldCnt2                     ' vga screen blank countdown
  long Countdown2                  ' vga screen blank countdown
  long Nickname                    ' single Long with # and 3 caps, eg #EPS for epsilon
  long NicknameBuffer              ' Store 4 bytes coming in from Port2
  byte Port2ControlFlag            ' flag to say being controlled by port2
  word ScreenColor                 ' RRGGBB00_RRGGBB00 background/foreground
  word buffcog              ' location of serial port cog code (for recycling) replaces buff[] array
  word buffdisk                   ' recycles cog code space, replaces disk_buff array
  byte buffbyte                   ' byte for buffcog

  'Parameter list for CPU objects, do not change order.
  long reg_base                   'Address (in HUB)of Z80 register file
  long io_base                    'Address (in prop RAM) of the I/O contol interface

  'word err                        'Error code return value
  long err                        'Error code return value

  long switch_reg
  long simhcmd

'For SRAM CPM Disk Loading
  long  filename
  long  srcname
  byte  buffer[16]
  long  sectorsize
'For microSD routines

  long  ioControl[2]               ' SD parameters

  long  blockno                    ' blockno within file
  long  sdsector                   ' sector no
  long  count                      ' count of chars to read
  'byte  buff[256]                  ' save some memory use 256 bytes only not used, now use buffcog


  byte hdsk_reset_count            'Counts number of consequiteve hdsk reset commands.
  byte hdsk_current_command        'The hdsk command currently in progress
  byte hdsk_last_command           'The last complete hdsk command issued
  byte hdsk_command_pos            'Counts through the outputed command parameter bytes (OUTs)
  byte hdsk_reply_pos              'Counts hdisk port reply bytes (INs)
  byte hdsk_drive                  'The hdsk drive to read/write with this command
  byte hdsk_sector                 'The hdsk sector to read/write with this command
  byte filler_1
  word hdsk_track                  'The hdsk track to read/write with this command
  word hdsk_dma                    'The DMA address to read/write from with this command
  long hdsk_dpb_ptr                'Pointer into DPB tables.


  long drive_base[18]              '// SD base address of drives 0..15 (A:..L:) + boot code
  long sector_number               'Sector of a disk starting from 0
  long sd_block_number             'SD card 512 byte block number starting from 0
  long block_offset                'Byte offset of a sector within a block
  long current_block_number        'The SD block currently in the disk buffer
  'byte disk_buff[512]              'The disk I/O buffer holds one SD card block.
'  long xxxx                        'required for overflow of disk_buff in floppy mode !!!



OBJ

'  cpu           : "Zicog_Drac"               ' Zilog Z80 emulator in one COG by heater
  cpu           : "zicogLMM"
'  cpu           : "yaz80"                 ' LMM driver March 2010
  mem           : "fsrwFemto"                ' sd card driver and RAM access
  vgatext       : "VGA_1024_VT100"           ' VGA VT100 Terminal Driver
  kb            : "keyboard"                 ' keyboard
  'RamLatches    : "Dracblade"                ' to talk to the ram chip
    ' RAM access is now through the modified SD card cog see new object "mem" above
  LCD           : "DracLCD"                  ' LCD 20x4 driver
  UART          : "pcFullDuplexSerial4FC"    ' 4 port serial driver in one cog.
  Delay         : "Timing"                   ' delay routines

DAT

'The Z80 Registers.
'Do not change the order of these.
'Registers are defined in pairs as words so as to enable them to be
'accessed individually as BYTE or in pairs as WORD.
'The ordering makes it easier to find a register specified by a 3-bit
'register field within an op code or to swap between main and
'alternate register sets. 
register_file word
bc_reg        word      0
de_reg        word      0
hl_reg        word      0
af_reg        word      0
bc_reg_alt    word      0 
de_reg_alt    word      0
hl_reg_alt    word      0
af_reg_alt    word      0
im_reg        word      0
ix_reg        word      0
iy_reg        word      0
sp_reg        word      0
pc_reg        word      0
  
'IO interface area. Do not change the order of these.
io_cmd_block  long
io_command    long      0
io_port       long      0            
io_data       long      0


'Hard Disk Parameter Blocks
'Each DPB is 17 bytes as defined for DPBs in CP/M followed by 2 bytes sector size as used by SIMH
hdsk_dpb
dpb_0                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)
                                'N.B. SS must be 128 for CP/M 2 can be varied for CP/M 3 hard disks.
{
' commented out all other drive types as only ever use SD 8mb drive images (drive type 0)
dpb_1                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determdirine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)
                                'N.B. SS must be 128 for CP/M 2 can be varied for CP/M 3 hard disks.

dpb_2                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)

dpb_3                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)

dpb_4                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)

dpb_5                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)

dpb_6                           '1MByte Altair mits2 floppy as HDSK params (no skew)
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $04     'data allocation Block SHift factor
:blm          BYTE      $0F     'data allocation block mask
:exm          BYTE      $00     'EXtent Mask
:dsm_low      BYTE      $EF     'maximum data block number (low_byte)
:dsm_high     BYTE      $01     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $00     'total number of directory entries (high byte)
:al0          BYTE      $F0     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)
                                'N.B. SS must be 128 for CP/M 2 can be varied for CP/M 3 hard disks.

dpb_7                           '440K RAM disk as HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $04     'data allocation Block SHift factor
:blm          BYTE      $0F     'data allocation block mask
:exm          BYTE      $00     'EXtent Mask
:dsm_low      BYTE      $DF     'maximum data block number (low_byte)  $006F=111*32*128=454,656=444KB <======??????
:dsm_high     BYTE      $00     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $00     'total number of directory entries (high byte)
:al0          BYTE      $F0     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $00     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)
                                'N.B. SS must be 128 for CP/M 2 can be varied for CP/M 3 hard disks.
}
PUB start | i
'   delay.pause1s(1)              ' short startup delay if required
   current_block_number := -1    'No SD block read into buffer yet so set invalid value
   TerminalFlag :=0
   Port2ControlFlag :=0                                 ' not being remotely controlled (255=yes)
   UART.init                                            ' initialise the uart
   UART.AddPort(0,31,30,-1,-1,0,0,baud)                 ' set up port 0
   UART.AddPort(1,25,24,-1,-1,0,0,1200)                 ' set up port 1 baud 1200
   UART.Start                                           ' start the cog
   termcolor:=5
   curset := 5
   vgatext.start(video)                                 ' start the vga driver
   vgatext.startcursor
' background and foreground colors - select one of these and comment the rest out
  screencolor := %00001000_11111100                   ' Blue_White  blue =3/4 power 10, white =full
'  screencolor := %00001000_11110000                   ' Blue_Gold
'  screencolor := %00000000_11111100                   ' Black_White
'   screencolor := %00000000_00110100                   ' Black_Green + tiny amount of blue
'  screencolor := %00000000_11100000                   ' Black_Orange
   vgatext.color(screencolor)                          ' %RRGGBB00 (4 levels each)
   vgatext.cursorset(curset)
   vgatext.clearscreen                                  ' clear the screen
   kb.startx(26, 27, NUM, RepeatRate)                  'Start Keyboard Driver
   LCD.start                                            ' start the LCD display
   PrintStringCR(string("Zicog with Dracblade"))        ' send to three places
' note the following two lines can't be used on the serial port nor the vga driver as these are reloaded (eg change baud rate and screensaver)
   buffcog :=kb.location                       ' find where the kb port cog code was - must run AFTER cog started
   buffdisk := buffcog+256                        ' location of temp disk buffer - replaces disk_buff

   reg_base        := @register_file                    ' find location in compiled code of register_file
   io_base         := @io_cmd_block                     ' find location in compiled code of io_cmd_block
   switch_reg      := $08                                '$08 is the switch setting SIMH recommends for 4k Basic etc to run
   StartCountdown1                                      ' start a countdown timer for sd card
   StartCountdown2                                      ' start countdown timer for vga blanking
'------------------------------------------------------------------------------
' Start the disks (from SD card)
' Load SRAM with Z80 RAM (64KB) & optional CPM Ram Disk (448KB/960KB)
'------------------------------------------------------------------------------
  blockno := 0
  count := 8192
  FindSDblock                                           'set the SD card file/drive base(s)
'  PrintStringCR(string("Loading SRAM"))
  LoadSRAM
'------------------------------------------------------------------------------
' ZiCog Emulation starts here...
'------------------------------------------------------------------------------
{
#ifdef CPU_Z80
  PrintStringCR(string("Z80"))
#else
  PrintStringCR(string("8080"))
#endif
}
  err := \cpu.start(@reg_base)
'  err :=\cpu.start(0,@reg_base,0,0)
  DisplayError(not err)
  'Start the Altair I/O simulation
  ReadMyName
  CreateNickname

  Service_IO

PRI FindSDblock | i, n, r                               'read blockno of file for count into buffer

  r := mem.start(@ioControl)                           'start the SD routines
  CheckError(not r)
  PrintStringCR(string("Mount FAT16 SD with .DSK files"))
  r := \mem.mount(spiDO,spiClk,spiDI,spiCS)      'mount the SD
  CheckError(r)

  'HDisks A:,B:,C:,D:,E:,F:,G:,H:
  FindDriveBase(0, string("A.DSK"))              'set the SD card drive base(s)  8MB CPM2.2
  FindDriveBase(1, string("B.DSK"))              'set the SD card drive base(s)  8MB
  FindDriveBase(2, string("C.DSK"))              'set the SD card drive base(s)  8MB Dr_A xmodem+
  FindDriveBase(3, string("D.DSK"))              'set the SD card drive base(s)  8MB
  FindDriveBase(4, string("E.DSK"))              'set the SD card drive base(s)  8MB vedit & pascal
  FindDriveBase(5, string("F.DSK"))              'set the SD card drive base(s)  8MB
  FindDriveBase(6, string("G.DSK"))              'set the SD card drive base(s)  1MB equiv floppy
  FindDriveBase(7, string("H.DSK"))              'set the SD card drive base(s)  <1MB RamDisk
  'other files Q:,R:
  FindDriveBase(17,string("BOOT.DSK"))              'set the SD card drive base(s)    $FF00 for boot hdisk

  r := mem.stopSDCard                                    'stop SD card, but do not stop cog
  CheckError(r)


PRI FindDriveBase(n, st) | r, fsize                     ' set the SD drive_base for the drive no
  PrintChar($41 + n)                                    ' print the names of the disk files
  PrintChar("=")
  PrintStringCR(st)

  r := mem.popen(st,"r")                                 'Open file for read (blank image)
  CheckError(r)

  'r := sd.pread(@buff,1)                                'get a char to find the sector address ****
  r := mem.pread(buffcog,1)                                'get a char to find the sector address ****
  CheckError(r)
  drive_base[n] := ioControl[0] - 1                     'get the sectorno of the file (returns the sector+1)

PRI LoadSRAM |  r 'i, j, r, b1, b2
'  count:=256                                           ' save some memory using 256 bytes max

'  b1 := cpmsize / count                                 '8KB blocks in 64KB CPM RAM
'  b2 := sramsize / count                                '8KB blocks in 512KB/1MB SRAM

  ' this is now done in the combined sd card/RAM driver
  'r := RamLatches.start                                 'start the dracblade driver
'  CheckError(r)
  HighLatchZero ' high latch byte to zero
  ' could use any bank here and store it in highlatchbyte 00000000_00000nnn_00000000_00000000
  'buff[0] := $C3                                        'jmp to $FF00 at $0000 (faster)
  'buff[1] := $00
  'buff[2] := $FF
    'bytefill(buffcog+0,$C3,1) ' this stores a byte
    'bytefill(buffcog+1,$00,1) ' this stores a byte
    'bytefill(buffcog+2,$FF,1) ' this stores a byte
    'longfill(buffcog,$00FF00C3,1) ' line below is even better!
    long[buffcog]:=$00FF00C3 ' program the jump in one line of spin.
  'vgatext.dec(getbuff(0)) ' this is the same as the old buff[] array
  'RamLatches.DoCmd("W", @buff, 0, 3) ' write 3 bytes in buff array to address 0
  mem.DoCmd("W", buffcog, 0, 3) ' write 3 bytes in buffcog to address 0
  ' little debugging routines to see if can read a byte back from the sram
  'vgatext.hex(buff[0],2) ' this should print C3
  'buff[0]:=1 ' this changes it to 01
  'vgatext.hex(buff[0],2) ' this should print 01
  'RamLatches.DoCmd("R", @buff, 0, 3) ' read 3 bytes in buff array to address 0
  'vgatext.hex(buff[0],2) ' this should print C3 again

  Countdown1:=0                                         ' value to tell sd card to initialise
  CheckSDCard                                            ' init sd card
  'r := sd.readSDCard(drive_base[17], @buff, 256)        'get 256 bytes "BOOTxxxx.ROM"
  r := mem.readSDCard(drive_base[17], buffcog, 256)        'get 256 bytes "BOOTxxxx.ROM"
  CheckError(r)
  'PrintStringCR(string("Boot code from SD card"))
  'r := RamLatches.DoCmd("W", @buff, $FF00, 256)               'write 256 bytes from buff[] to sram
  r := mem.DoCmd("W", buffcog, $FF00, 256)               'write 256 bytes from buff to sram
  CheckError(r)

PRI CheckError(r)
  if r < 0
    PrintStringCR(string("Er1"))
'    UART.str(0,string("***Error*** $"))              ' uncomment to see the error
'    UART.hex(0,r,2)
'    UART.str(0,string(" ***HALTED***",13,10))
    mem.stop
    repeat                                              ' Loop here forever

PRI DisplayError(errx)
  if errx
    PrintStringCR(string("Er2"))
  else
    PrintStringCR(string("Loading Zicog"))

'------------------------------------------------------------------------------
' Service I/O
'------------------------------------------------------------------------------
PRI service_io | d ' d=delay
'Altair I/O simulation service loop
'I/O control block commands
'io_cmd_out      = $01
'io_cmd_in       = $02
'io_cmd_break    = $03
  repeat
    CheckCountdown1                                     ' sd card timeout
    CheckCountdown2                                     ' vga blank timeout
    TestPort2                  ' eg #ALP or #BET, #OFF, #WHO #BYE
    'vgatext.str(string("serviceio"))
    repeat 250 ' spend most of the time checking i/o don't make this too high otherwise the screensaver doesn't work (not sure why, must be something in the subtraction math)
      case io_command
        io_cmd_break:                                     ' 03
          on_break                                        ' Break (for single stepping)
        io_cmd_in:                                        ' 02
          on_input                                        'An input operation
        io_cmd_out:                                       ' 01
          on_output                                      'An output operation


PRI on_break
'Handle CPU break condition eg for single stepping
  print_regs
  dumpPC                                                ' print next few hex bytes
  'PrintStringCR(string("Unimplemented instruction one before location PC, any key to continue"))
  printstringcr(string("Er3"))
'  repeat while UART.rx(0) <> $20                   ' wait for 'space' from terminal
  repeat until kb.getkey<>0                         ' wait for local keypress
  io_command := 0


PRI on_input
'Handle Z80 IO port reads
  case io_port
'    srport:                                             'Altair front panel switches register
'      in_srport
    simhport:                                           'SIMH control port
      in_simhport          
    constat:                                            'Console status port $0F
      in_constat
    condata:                                            'Console data port
      in_condata
    n8vem_uart_test:                                    ' test port $6D
      in_test_n8vem                                     ' is a byte ready returns %00100000 for no byte, %00100001 for a byte
    n8vem_uart_in:                                      ' ask for input from port $68
      in_uart_n8vem                                     ' get the byte
    port2_test:                                         ' test second port $7D
      in_test_port2                                     ' is a byte ready returns %00100000 for no byte, %00100001 for a byte
    port2_io:                                           ' get the byte from port $78
      in_port2                                          ' get the byte
    hdskport:
      in_hdskport                                       'SIMH style hard disk port
    keystat:                                            ' test just the keyboard (constat above is a combo of both ports and keyboard)
      in_keystat
    keydata:                                            ' get the keyboard character
      in_keydata
'    ram:                                                ' read a byte from ram (port $76)
'      peek_ram                                          ' this works but commented out as takes a lot of memory
' for some reason the following reader code cannot sit underneath constat/condata above, it only works down here otherwise xmodem won't transfer files?!
    punstat:                                            ' punch/reader SIO combined on SIMH
      'vgatext.str(string("reader status"))
      in_readerstat
    pundata:                                           ' punch/reader SIO combined on SIMH
      'vgatext.str(string("reader byte"))
      in_readerdata
    other:
      PrintStringCR(string("Er14"))  'comment these out if certain no more ports being used
      'vgatext.dec(io_port)
      io_data := 0
  io_command := 0                                       'Set flag ready for next io

PRI in_test_n8vem
'  for xmodem tests if any input from serial port 0
' input present = bit 0. Output ready to go = bit 5 (always high). bit1=0 (carrier detect)
  if UART.rxavail(0)
    io_data :=%00100001                                 'yes
  else
    io_data :=%00100000                                 'no


PRI in_uart_n8vem                                          ' for xmodem transfers
  if UART.rxavail(0)
    io_data := UART.rx(0)                                  ' get the byte
  else
    io_data := "?"                                      '<=== at least show a character to indicate a problem

'PRI in_srport
''Handle reads from the Altair front panel switch port
'  io_data := switch_reg
'  PrintStringCR(string("SR port read"))


PRI in_simhport
'Handle reads from the SIMH simulator status port
  if simhcmd == hasbankedmemory
    io_data :=  1
    simhcmd := 0

PRI in_constat
'Handle reads from the console status port
  Keyflag :=kb.Gotkey                                   ' 255=true (character), 0=false is there a key in the buffer
  UART0Flag:= UART.rxavail(0)                           ' 255=true (character), 0=false
  if Port2ControlFlag                                   ' port2 switched on remotely
    UART1Flag := UART.rxavail(1)                        ' there is a byte in port2 (uart1)
  if KeyFlag or UART0Flag or UART1Flag                  ' something has come in from one of the input sources
    io_data := %00000011                                'Return flag to say a character is available
    CheckScreensaver                                    ' keep the vga display on if keypresses
  else
    io_data := %00000010                                'No: Return no char ready for input

PRI in_condata
'Handle reads from the console data port, expects flags to be set correctly
    if UART0Flag                                          ' check uart first
      io_data := UART.rxcheck(0)                          ' Yes: Return a character
      UART0Flag :=0                                       ' reset the flag
    else                                                  ' if all cleared then local keyboard
      if KeyFlag                                          ' is there a keyboard char?
        key :=kb.key                                      ' get the character
        ProcessKey                                        ' check for some high characters and remap
        io_data :=key                                     ' send to CP/M
        Keyflag :=0                                       ' reset the flag
      else
        if UART1Flag
          io_data := UART.rxcheck(1)                          ' Yes: Return a character (port2)
          UART1Flag :=0                                       ' reset the flag

PRI in_readerstat
  'vgatext.str(string("punch status"))
  if UART.rxavail(1)
    io_data := %00000011
  else
    io_data := %00000010
' if outputting to the punch device first checks if can send this. This should always be true ie bit 1
' if inputting from the reader device then bit 0 is true or false depending on whether data is available

PRI in_readerdata
  'vgatext.str(string("punch data input"))
  io_data := UART.rx(1) ' get byte from port2 (ie uart 1), assumes data is there from punstat

PRI in_keystat  ' get just the keyboard, see also in_constat
    io_data:= kb.Gotkey                                 ' 0 = no key read, 255= true, got a character

PRI in_keydata ' get the keyboard character, assumes one is there as above has been tested
    key :=kb.key                ' dedicated keyboard input on a port
    Processkey
    io_data :=key
    checkscreensaver            ' keep screen on if keys being pressed

PRI hdsk_check_dts : status
'Check that selected hdsk drive, sector and track are within limits
  if (hdsk_drive < nhdisks) and (hdsk_track =< 2048) and (hdsk_sector =< 128)
    status := $00
  else
    PrintStringCR(string("Er4"))
    DisplayHDiskDTS
    status := $FF


PRI refresh_hd_cache | r
'Ensure that the sector number in use is in the current SD block in memory.
'If not load that block to disk buff from SD card.
  sector_number := (hdsk_track * 32) + hdsk_sector    'Find correct block for the selected track and sector                        '..and sector
  sd_block_number := sector_number >> 2               'Determine the SD block of this sector..
  sd_block_number += drive_base[hdsk_drive]           'get the first sector of the selected drive
  block_offset := (sector_number & %11) * 128         'Determine the position of this sector within the block
  if sd_block_number <> current_block_number          'If the block is not in disk buffer then load it
  CheckSDCard                                         ' restart if has timed out
  'err := \sd.readSDCard(sd_block_number, @disk_buff, 512)     'read the block! (complete 512 bytes)
  err := \mem.readSDCard(sd_block_number, buffdisk, 512)     'read the block! (complete 512 bytes)
    if err
      PrintStringCR(string("Er5"))
      DisplayError(err)
      DisplayHDiskDTS
      io_data := $FF                                  'Return error code
    else
      current_block_number := sd_block_number         'Make this new block current


PRI in_hdskport_read | r
'Handle read from hdskport during a disk read command
'This read triggers the actual disk sector transfer and must return an error code (To the Z80 not the calling Spin)
'N.B. sectors are mapped to blocks 4 to 1
  if hdsk_check_dts                                   'Check drive, track and sector are sane
    io_data := $FF                                    'Return error if not
    return
  refresh_hd_cache
  'r := RamLatches.DoCmd("W", @disk_buff + block_offset, hdsk_dma, 128) 'Virtual "DMA" the sector into DMA buffer in SRAM
  r := mem.DoCmd("W", buffdisk + block_offset, hdsk_dma, 128) 'Virtual "DMA" the sector into DMA buffer in SRAM
  CheckError(r)
  io_data := $00                                        ' Return OK.
  hdsk_last_command := hdskNull                         ' Command completed

PRI in_hdskport_write | r
'Handle read from hdskport during a disk write command
'This read triggers the actual disk sector transfer and must return an error code(To the Z80 not the calling Spin)
'N.B. sectors are mapped to blocks 4 to 1
  if hdsk_check_dts                                     'Check drive, track and sector are sane
    io_data := $FF                                      'Return error if not
    return

  refresh_hd_cache

  'Virtual "DMA" the sector from DMA buffer address in Z80 RAM
  'read the block in from Z80 sram
  'r := RamLatches.DoCmd("R", @disk_buff + block_offset, hdsk_dma, 128) 'Virtual "DMA" the sector from DMA buffer from SRAM
  r := mem.DoCmd("R", buffdisk + block_offset, hdsk_dma, 128) 'Virtual "DMA" the sector from DMA buffer from SRAM
  CheckError(r)
  checkSDCard                                           ' restart if timed out
  'err := \sd.writeSDCard(sd_block_number, @disk_buff, 512)   'write the block! (complete 512 bytes)
  err := \mem.writeSDCard(sd_block_number, buffdisk, 512)   'write the block! (complete 512 bytes)
  CheckError(err)
  if err
    PrintStringCR(string("Er6"))
    DisplayError(err)
    DisplayHDiskDTS
    io_data := $FF                                      'Return error code
  else
    io_data := $00                                      'Return OK.
  hdsk_last_command := hdskNull                         'Command completed


PRI in_hdskport_param
'Handle read from hdskport during disk parameters command
'Simply return the next param byte from a DPB.
  io_data := hdsk_dpb[hdsk_reply_pos++]
  io_data := BYTE[hdsk_dpb_ptr++]
  hdsk_reply_pos++
  if hdsk_reply_pos == 19
    hdsk_reply_pos := 0
    hdsk_last_command := hdskNull                   'Command completed
    'PrintStringCR(string("Hdskport"))

PRI in_hdskport
'Handle reads from the hard disk interface port
'Handling is dispatched to an appropriate function depending on the disk operation currently in progress
  case hdsk_last_command
    hdskRead:
      in_hdskport_read
    hdskWrite:
      in_hdskport_write
    hdskParam:                                          'Current command is get disk params so just return successive parameter bytes from DPB.
      in_hdskport_param
    other:
      PrintStringCR(string("Er7"))

PRI on_output
'Handle Z80 IO port writes
'simhport        = $FE
'hdskport        = $fd                   ' control port for simulated hard disk  
'vgatext.dec(io_port)
  case io_port
'    ledport:                                            'An 8 bit parallel output with LEDs attached
'      out_ledport
    simhport:                                           'SIMH control port $FE
'      out_simhport
    condata:                                            'Console data port $11
      out_condata
    constat:                                            'Console status port
'      out_constat
    punstat:
                                                        ' do nothing outputting the status
    pundata:                                            ' connect CP/M punch to port 2
      out_port2                                         ' output to port 2 = the punch device
    n8vem_uart_out:                                     '&H68
      out_n8vem_byte
    port2_io:                                            ' for output to port 2
      out_port2
    port2_baud:                                         ' set baud rate for port 2
      baud_port2
    Forecolor:                                          ' set the foreground color RRGGBB00
      set_forecolor
    Backcolor:                                          ' set background color RRGGBB00
      set_backcolor
    hdskport:                                           ' $FD
      out_hdskport                                      'SIMH style hard disk port
    myname_port:                                        ' port &H70
      GetMyname                                         ' returns board name in DMA 80H
    Port2_Off:                                           ' port &H71
      Port2_Disconnect                                   ' disconnect port 2 control
'    ram:                                                ' write a byte to ram (port $76)
'      poke_ram                                          ' commented out as not used and takes up spin code space

    other:
      PrintStringCR(string("Er8"))
      'UART.hex(0,io_port,2)                               'print the port number for debugging
  io_command := 0                                       'Set flag ready for next io

'PRI out_ledport
'Handle writes to the LED port
'No, SIMH never had this. I added it for the Propeller demo board.
  'outa[16..23] := io_data                              'Output the port data  (if re-enabled do #ifdef)
'  UART.str(0,string("LED port = "))
'  UART.hex(0,io_data,2)



'PRI out_simhport
'Handle writes to the SIMH simulator control port
'   uart.str(string("Out_Simhport "))
'   crlf

PRI out_condata   'Handle writes to the console data output port
    PrintChar(io_data)                                   ' sends to terminal, LCD and vga

'PRI out_constat
'Handle writes to the console command port
'  if io_data == creset
    'UART.str(string("Console reset"))
    'crlf
'  else
'    PrintStringCR(string("Unknown console write"))
'    UART.hex(0,io_data,2)          ' send to terminal only

'PRI out_punstat
'Handle writes to the punch device command port
'  if io_data == creset
    'UART.str(string("Punch reset"))
    'crlf
'  else
'    PrintStringCR(string("Unknown punch write"))
'    UART.hex(0,io_data,2)          ' send to terminal only


PRI out_n8vem_byte
  UART.tx(0,io_data)   ' send out the data to the port (for xmodem)
'  vgatext.str(string("send:")) ' for debugging
'  vgatext.dec(io_data)         ' for debugging
PRI hdsk_get_rw_params
  case hdsk_command_pos
    1:
      hdsk_drive := io_data                             '1st byte is the drive number
      hdsk_command_pos++
    2:
      hdsk_sector := io_data                            '2nd byte is the sector number
      hdsk_command_pos++
    3:
      hdsk_track := io_data                             '3rd byte is LOW of the track number
      hdsk_command_pos++
    4:
      hdsk_track := hdsk_track + (io_data << 8)         '4th byte is HIGH of the track number
      hdsk_command_pos++
    5:
      hdsk_dma := io_data                               '5th byte is LOW of the DMA address
      hdsk_command_pos++
    6:
      hdsk_dma := hdsk_dma + (io_data << 8)             '6th (and last) byte is HIGH of the DMA address
      hdsk_command_pos++
      hdsk_last_command := hdsk_current_command         'OUT's for current command completed
      hdsk_current_command := hdskNull
    other:
      PrintStringCR(string("Er9"))
      repeat


PRI out_hdskport_param
'Handle writes to the hdskport when get params command is in progress
  case hdsk_command_pos
    1:
      hdsk_drive := io_data
      hdsk_command_pos++
      hdsk_reply_pos := 0
      'hdsk_dpb_ptr := @hdsk_dpb + (hdsk_drive * 19)     'Point to correct DPB for this drive
      hdsk_dpb_ptr := @hdsk_dpb ' always point to dpb 0 for the dracblade, this is the simh 8mb file standard
      'vgatext.dec(hdsk_drive) ' so can see drive number 0-7
      hdsk_last_command := hdsk_current_command         'OUT's for current command completed
      hdsk_current_command := hdskNull
    other:
      PrintStringCR(string("Er10"))
      repeat


PRI out_hdskport_null
'Handle writes to the hdskport when no command is in progress
'io_data should be a new command byte.
  case io_data
    hdskRead:
      hdsk_current_command := hdskRead
      hdsk_command_pos := 1
    hdskWrite:
      hdsk_current_command := hdskWrite
      hdsk_command_pos := 1
    hdskParam:
      hdsk_current_command := hdskParam
      hdsk_command_pos := 1
    hdskReset:
      hdsk_current_command := hdskNull
      hdsk_last_command := hdskNull
      hdsk_command_pos := 1
      hdsk_reply_pos := 0
    other:
      PrintStringCR(string("Er11"))
      repeat


PRI out_hdskport
'Handle writes to the SIMH style hard disk port
'Handling is dispatched to an apropriate function depending
'on the command currently in progress.
'FIXME: RESET should be checked on every write no matter the command in progress.
'hdskNull        = 0                     ' No command
'hdskReset       = 1                     ' command to reset controller
'hdskRead        = 2                     ' read command
'hdskWrite       = 3                     ' write command
'hdskParam       = 4                     ' Get HD parameters command
' can't put diag messages in a case command, only at the beginning
  case hdsk_current_command
    hdskRead:
      hdsk_get_rw_params
    hdskWrite:
      hdsk_get_rw_params
    hdskParam:
      out_hdskport_param
    hdskReset:
      PrintStringCR(string("Er12"))
      repeat
    hdskNull:
      out_hdskport_null
    other:
      PrintStringCR(string("Er13"))
      repeat


PRI DisplayHDiskDTS              ' send to terminal only
  UART.str(0,string("D = "))
  UART.hex(0,hdsk_drive, 2)
  UART.str(0,string(" T = "))
  UART.hex(0,hdsk_track, 4)
  UART.str(0,string(" S = "))
  UART.hex(0,hdsk_sector, 2)
  UART.str(0,string(" DMA = "))
  UART.hex(0,hdsk_dma, 4)

PRI print_regs
'Print the Z80 registers
  RegString(string("AF ="))
  HexOutput(af_reg)
  RegString(string(" BC ="))
  HexOutput(bc_reg)
  RegString(string(" DE ="))
  HexOutput(de_reg)
  RegString(string(" HL ="))
  HexOutput(hl_reg)
  RegString(string(" SP ="))
  HexOutput(sp_reg)
  RegString(string(" PC ="))
  HexOutput(pc_reg)
  RegString(string(13,10))     ' new line for alternate register
  RegString(string("AF'="))
  HexOutput(af_reg_alt)
  RegString(string(" BC'="))
  HexOutput(bc_reg_alt)
  RegString(string(" DE'="))
  HexOutput(de_reg_alt)
  RegString(string(" HL'="))
  HexOutput(hl_reg_alt)
  RegString(string(" IX ="))
  HexOutput(ix_reg)
  RegString(string(" IY ="))
  HexOutput(iy_reg)
  RegString(string(13,10))

PRI DumpPC | i ' dump the surroundnig bytes in ram for single stepping
  'RamLatches.DoCmd("R", @buff, pc_reg-5, 16) ' read bytes from memory
  mem.DoCmd("R", buffcog, pc_reg-5, 16) ' read bytes from memory
  repeat i from 0 to 10
    vgatext.hex(i+pc_reg-5,4)             ' address then data byte
    vgatext.str(string(" "))
    'vgatext.hex(buff[i],2)
    vgatext.hex(getbuff(i),2)
    vgatext.str(string(13,10))

PUB HexOutput(HexValue)' send hex value to vga and to terminal
    UART.hex(0,HexValue, 4)
    vgatext.hex(HexValue,4)

PUB RegString(RString)
     UART.str(0,Rstring)                                ' send to the serial port
     vgatext.str(Rstring)                               ' send to vga display


PUB ProcessKey     ' pass key and changes a few values eg 200 is actually 08 backspace
    'vgatext.dec(key)                             ' for testing
    ' 222 is caps lock
    if key >576
       if key <603
          key:=key-576
    if key > 608  and key < 635                   ' Is it a control character?
       key:=key-608
    if key == 200                                 ' backspace
       key:=08                                    ' send backspace to CP/M
    if key == 203                                 ' Is it upper code for ESC key?
       key:= 27
    if key == 201                                 ' is it delete
       key:=127

PUB PrintStringCR(pstring)' sends pstring to the uart, to the LCD display and to the vga display
     UART.str(0,pstring)                                ' send to the serial port
     UART.str(0,string(13,10))                          ' add carriage return
     LCD.str(pstring)                                   ' send to 20x4 LCD
     LCD.str(string(13,10))                             ' add carriage return
     vgatext.str(pstring)                               ' send to vga display
     vgatext.str(string(13,10))                         ' add carriage return

PUB PrintChar(c) ' sends c to the uart, LCD display and the vga display
    UART.tx(0,c)                                  ' send to the PC
    LCD.out(c)                                    ' send to LCD display
    vgatext.txVT100(c)                            ' send to vt100 terminal
    if Port2ControlFlag                           ' if comms enabled for port 2
      UART.tx(1,c)                                ' send to this port as well
' -------------------------------------------------------------
' ************** port 2 routines ********************************
' note on the tx no need to check if ready as it is always ready (doesn't return control to zicog till has send out some bytes if buffer is full)
PRI out_port2               ' send to port 2 (male D9)
   UART.tx(1,io_data)        ' the uart driver numbers are 0,1,2,3 but we call this port 2 (ie goes to uart 1)
   ' 10 A=65                   ' mbasic example
   ' 20 OUT (&H78),A
PRI baud_port2 | baudrate
   ' send io_data multiple of 1200 eg 1=1200, 16=19200, 96=115200
   ' port number 7C
   baudrate:=io_data*1200
   UART.init                                            ' reinitialise the uart
   UART.AddPort(0,31,30,-1,-1,0,0,baud)                 ' set up port 0 again to 19200
   UART.AddPort(1,25,24,-1,-1,0,0,baudrate)             ' new baud rate
   UART.Start                                           ' restart the uart driver
   ' 10 BAUD = 1200                                     ' mbasic example
   ' 20 OUT (&H7C),BAUD/1200

PRI in_test_port2 ' port $H7D
' tests if any input from the serial port
' input present = bit 0. Output ready to go = bit 5 (always high). bit1=0 (carrier detect)
  if UART.rxavail(1)' uart port 1 is what I call port 2. Anything available?
    io_data :=%00100001                                 'yes
  else
    io_data :=%00100000                                 'no
' 10 A=INP(&H7D)                                        ' mbasic example
' returns either 32=no byte, or 33 which means there is a byte available
PRI in_port2 ' get the byte from port 2 = port &H78
'  vgatext.dec(uart.lookahead(1))                        ' a long with the next 4 bytes in the buffer
  if UART.rxavail(1) ' should always be true
    io_data := UART.rx(1)                                  ' get the byte
  else
    io_data := "?"                                      '<=== at least show a character to indicate a problem
' 10 A=INP(&H78)                                        ' mbasic example
' test program in mbasic to use port 2. Run a terminal program with a null modem (cross pins 2 and 3) at 19200 baud
'10 PRINT "Testing output character A to port 2 (use terminal program)"
'20 A=65
'30 OUT (&H78),A
'40 PRINT "Now testing for input. Hangs and waits for a keypress on terminal"
'50 A=INP(&H7D)
'60 IF A=32 THEN GOTO 50
'70 A=INP(&H78)
'80 PRINT "Character received ";
'90 PRINT CHR$(A)

PRI TestPort2 | i ' look for a message eg #ALP on port 2 that opens a radio comms channel to this board
   if (uart.lookbehind(1) == nickname) and (Port2ControlFlag == 0) ' look behind the head works better than looking forward from the tail. Don't have to wait 64 bytes if get out of synch
'     vgatext.str(string("Open Port2"))
'     uart.tx(1,"#")   ' send one byte back as an acknowledge (no need for long text strings just uses up a lot of memory
     uart.rxflush(1) ' clear the buffer
     Port2ControlFlag := 255                 ' port2 now sends bytes to the terminal
 ' check to see if a string #WHO has arrived 2357484F = 592922703
   if (uart.lookbehind(1) == 592922703)
     'vgatext.str(string("#WHO"))
     ' note there is no easy way to clear the buffer so it keeps reading the last 4 bytes
     ' over and over. So the transmitting program needs to send an #OFF
     ' do this with time delays. The Tx program waits 10 secs for a reply, and this spin
     ' program waits 15 seconds. Plus an initial random delay
     ' cnt >> 16 is 0 to 65535 so >>19 is to 8192
     delay.pause1ms(cnt>>19) ' random delay 0-8 seconds
     'vgatext.str(string("Reply"))
     repeat until i>7
       'vgatext.dec(i)
       uart.tx(1,myname[i])
       i++
     delay.pause1ms(15000) ' this must be longer than the sbasic program takes
     'vgatext.str(string("Delay finished"))
   ' check to see if a string #BYE has arrived $23425945 = decimal 591550789
   if (uart.lookbehind(1) == 591550789) and (Port2ControlFlag == 255)
     'vgatext.str(string("#BYE"))
     Port2_Disconnect
    ' check for global reset command #RES 23524553H = 592594259
   if (uart.lookbehind(1) == 592594259)
     start

PRI Port2_Disconnect ' called from above and also if screen has gone blank and timed out
'    vgatext.str(string("Close Port2"))
    Port2ControlFlag := 0                    ' set the flag to zero

' 10 A=0:' can be any value
' 20 OUT (&h71),A ' turns off control from port2
' or in assembly
'; File PORT2OFF.MAC Turn off control from Port 2
'; use zasm.sub with supersub
'; eg supersub zasm myfile    (where myfile is called myfile.mac but do not put the
';                             mac in the instruction - m80 knows it is implied)
'; zasm.sub is the following two lines
';M80 =$1 /Z/C/L/M
';L80 $1,$1/N/E
'
'START:		OUT	(71H),A			; send anything to port 71 to disable
'		RET				; back to CP/M
'		END

' *********** end port routines **********************************

'PUB WriteName ' test writing a name to the sd card from CP/M - call after the sd card initialised eg just before zicog
'  sd.popen(string("MYNAME.TXT",0),"w")         ' open file Myname.txt - terminate with 0
'  sd.pwrite(string("Alpha   "),8)  ' write 8 bytes to file, pad alpha beta gamma to 8 chars
'  sd.pclose ' close the file
{
' --------------------------------------------------------------------------------
' this all works but takes up spin space and might end up using the space for other things
PRI Peek_Ram | a ' read the value in a ram location
  ' pass address in DMA 80H=LSB, 81H=middle byte, 82H=MSB
  PeekPokeSetup
  a:=buff[1]<<8 | buff[0] ' 0 to 65535 combine two bytes
  RamLatches.DoCmd("R", @buff, a, 1) ' value to buff[0] ' need to use buffcog with no @
  io_data:=buff[0]
  HighLatchZero

PRI Poke_Ram | a ' read the value in a ram location
  ' pass address in DMA 80H=LSB, 81H=middle byte, 82H=MSB
  PeekPokeSetup
  a:=buff[1]<<8 | buff[0] ' 0 to 65535 combine two bytes
  buff[0]:=io_data ' put io_data in buff array ready to output
  RamLatches.DoCmd("W", @buff, a, 1) ' write io_data which is in buff[0] ' need to use buffcog with no @
  HighLatchZero

PRI PeekPokeSetup
  RamLatches.DoCmd("R", @buff, $80, 3) ' read 3 bytes from DMA area to buff array ' need to use buffcog with no @
  HighLatchByte :=buff[2] << 16 ' multiply by 65536 to get high byte
  RamLatches.DoCmd("H", 0, HighLatchByte, 0)   ' setup high byte
}' ------------------------------------------------------------------
PRI HighLatchZero
  HighLatchByte :=%00000000_00000000_00000000_00000000 ' xxxxxxxx_nnnnnnnn_xxxxxxxx_xxxxxxxx where n is A16 to A23
  mem.DoCmd("H", 0, HighLatchByte, 0)   ' restore for CP/M

{
10 REM poke then peek a value to ram (0 to 524287 ie to 2^19 minus 1)
20 ADDRESS!=2^19-1
30 PRINT ADDRESS!
40 REM numbers >32768 both bitwise AND, MOD and \ integer division don't work
50 GOSUB 120: REM send out the address
60 OUT &H76,5:REM test byte
70 GOSUB 120: REM resend address
80 A=INP(&H76)
90 PRINT A: REM should print out the value three lines up
100 END: REM end of program
110 REM **********************************************
120 REM pass ADDRESS! up to 2^19 and splits into 3 bytes and puts in DMA
130 BYTE!=INT(ADDRESS!/65536!)
140 POKE &H82,BYTE!
150 TEMP!=BYTE!*65536!
160 TEMP2!=ADDRESS!-TEMP!
170 BYTE!=INT(TEMP2!/256)
180 POKE &H81,BYTE!
190 TEMP!=BYTE!*256
200 TEMP2!=TEMP2!-TEMP!
210 BYTE!=TEMP2!
220 POKE &H80,BYTE!
230 RETURN
}
PUB ReadMyName | r
  ' MYNAME.TXT is a tiny text file max 8 bytes created with notepad and copied to the SD card
  ' names are ALPHA BETA GAMMA EPSILON etc through the Greek alphabet
  ' pad unneeded bytes with spaces

  r:=mem.popen(string("MYNAME.TXT"),"r")
  r:=mem.pread(@myname,8)                                  'get 8 bytes Alpha... Beta.... etc
  if r<0
    printstringcr(string("MYNAME.TXT not found"))
  else
    printstringcr(String("Board name = "))
    printstringcr(@myname)         ' terminates with 0 and the array is 9 bytes and only copied 8 to the array so the last byte is 0 which is the terminator
  mem.pclose

PUB GetMyname | r
    ' reads 8 bytes from MYNAME.TXT and puts them in the DMA starting at address 80
    ' could pass io_data here if wanted to pass a value (0-255)
    ' small program is myname.mac OUT (70H),A  where A doesn't really matter but is passed in io_data
  r:=mem.popen(string("MYNAME.TXT"),"r")
  r:=mem.pread(@myname,8)                                  'get 8 bytes Alpha... Beta.... etc
  if r => 0           ' *** note => not >= ***
    'vgatext.str(@myname)
    mem.DoCmd("W", @myname, 128, 8) ' move 8 bytes from myname array to address 80H
  mem.pclose
{
10 REM an out to port &H70, with any value to get back the name of the board
20 OUT &H70,0
30 FOR I=&H80 TO &H87
40 J=PEEK(I)
50 IF J>=65 AND J<127 THEN A$=A$+CHR$(J)
60 NEXT I
70 PRINT A$
80 PRINT LEN(A$)
}

PUB CreateNickname ' create a one long nickname eg Epsilon is #EPS  (capitals)
    nickname:=("#"<<24)+(myname[0]<<16)+(myname[1]<<8)+(myname[2])
    'vgatext.dec(nickname) ' #ALP should be 591481936

PUB StartCountDown1             ' used by sd card
   OldCnt1:=cnt
   Countdown1:=3000             ' 3 second delay

PUB CheckCountdown1 | r ' used for sd card timeout
   'vgatext.dec(countdown1)
   if Countdown1 <> 0
     if cnt > (Oldcnt1 + 20000000) ' add a big value as numbers are too tiny as Spin is too fast! Otherwise the calculated value below ends up zero most of the time
       'vgatext.dec(countdown1)
       'vgatext.str(string("Subtracting"))
       Countdown1:=Countdown1-((cnt-OldCnt1)/(clkfreq/1000)) ' subtract milliseconds since last check
       OldCnt1:=cnt  ' store the current clock counter for next time
       if Countdown1 =< 0  ' *** note the syntax is =<, not <= ***
         Countdown1:=0              ' for quick tests only run the countdown if CountDown1 <>0
         'vgatext.str(string("Finishing SD card"))
         r := mem.stopSDcard                                  'stop the SD (free up pins, but keep cog running)
         CheckError(r)
         mem.DoCmd("F",0,0,0)                         ' led off
PUB CheckSDCard | r ' if SD card has timed out then restart
    if Countdown1 == 0 ' if timed out
      'vgatext.str(string("Reinitialise SD card"))
      r := mem.initSDcard(spiDO,spiClk,spiDI,spiCS)        'init the SD - stopped if timeout
      CheckError(r)
      mem.DoCmd("N",0,0,0)                           ' led on
    StartCountdown1 ' restart the counter every time called

PUB StartCountDown2             ' used by vga blanking
   OldCnt2:=cnt
   Countdown2:=300000             ' number of milliseconds  eg 300000 for 5 minutes

PUB CheckCountdown2 ' for screensaver
   if Countdown2 <> 0
     if cnt > (Oldcnt2 + 20000000) ' add a big value as numbers are too tiny as Spin is too fast! Otherwise the calculated value below ends up zero most of the time
       Countdown2:=Countdown2-((cnt-OldCnt2)/(clkfreq/1000)) ' subtract milliseconds since last check
       OldCnt2:=cnt  ' store the current clock counter for next time
       'vgatext.dec(countdown2)
       'vgatext.str(string("  "))                               ' send to vga display
       'vgatext.dec(oldcnt2)
       'vgatext.str(string(13,10))                         ' add carriage return
       if Countdown2 =< 0  ' *** note the syntax is =<, not <= ***
         Countdown2:=0
         vgatext.stop                ' turn off the vga cogs
         Port2_Disconnect            ' disconnect port 2 as no input for a while

PUB CheckScreensaver   ' restart when a key pressed
   if Countdown2 == 0
     vgatext.start(video)     ' turn on the vga cogs
   StartCountdown2

PRI set_forecolor
   screencolor := (screencolor & $FF00) + io_data  ' preserve background
   vgatext.color(screencolor)
' mbasic OUT (&H72),&HE0 turns the screen to orange text 11100000 RRGGBB00 use windows calculator to convert binary to decimal

PRI set_backcolor
   screencolor := (screencolor & $FF) + (io_data <<8) ' preserve foreground
   vgatext.color(screencolor)
{
PUB Storebuff(i,j) ' replace the buff[] array with the leftover space from the serial driver
    ' commented out as can do this directly in one line and save some memory
    ' i=location (0-255),uses common variable buffbyte, moves 1 byte
    i:=i + buffcog
    'vgatext.dec(i)
    bytemove(i,@j,1) ' this stores the byte
    'buffbyte:=0 ' buffbyte
    'bytemove(@buffbyte,i,1) ' get the byte back  - this works
    'vgatext.dec(buffbyte) ' prints correct value
}
PUB GetBuff(i) ' i =location
    i:=i + buffcog
    bytemove(@buffbyte,i,1)
    return buffbyte

