'' System Version 2.003.rr001
'' loaderInit Changes:
''
''   2006/09/24 - Corrected action table for i2cRead0Cmd
''   2006/10/02 - Improved assembly comments.  Changed i2cReset, i2cStop
''   2006/11/02 - Changed data setup time.  Changed i2cReset timing
''   2006/11/03 - Changed read/write method speed to 100KHz
''   2006/11/04 - Added checkPresence, writeWait, and computeTimes methods
''   2006/11/06 - Limited boot loading to 32K - 16 (OS uses last 16 bytes)
''                Added ioVerifyCmd and verifyEEPROM method
''   2006/11/09 - Modified boot and verify to use the minimum of the actual
''                length of the program stored in vbase ($0008) or the
''                specified byte count.
''                Note also that these routines don't know about option bits.
''   2006/11/10 - Now control block address is passed to start routine
''   2007/01/09 - Added SPI routines for SD card FAT access
''   2007/01/13 - Corrected checksum when ioBoot or ioVerify
''   2007/01/13 - Stores stack marker & clears VAR area on ioBootCmd
''   2007/02/06 - Stores stack marker & clears VAR area on ioSpiBoot
''   2007/02/13 - Changed the way verify mode was done, combined code
''   2007/02/20 - Added ioStopLdr option to stop the loader's cog
''   2007/02/22 - Corrected bootSDCard.  Needs start and initSDCard calls first.
''   2009/04/02 - Cluso99: Fix to force the SD card to tristate DO in spiDoStop
''   2010/03/28 - hairymnstr: Removed i2c and booting, added RAM access for DracBlade

'' This portion is normally run only once during initialization and the driver remains
'' resident in a cog.  These routines can be used completely independently of the rest
'' of the Propeller OS.  The start routine here expects the address of a 2 long area
'' to be used for passing information to the I/O routines in the COG.  This area should
'' be located in an area of memory not expected to be overlaid by data or a program that
'' might be loaded since the COG routines will be accessing this information after an
'' operation has completed.


'' -------------------------------------------------------------------
'' |   cmd/status   |          I/O pin / device / address            |
'' -------------------------------------------------------------------
'' |           byte count           |          HUB address           |
'' -------------------------------------------------------------------

'' 27 Mar 2010: Frankenstien's Monster:
'' Major mod by Nathan Dumont (hairymnstr) to include RAM access functions for DracBlade
''
'' -------------------------------------------------------------------
'' |   cmd/status   |                 RAM Address                    |
'' -------------------------------------------------------------------
'' |          byte count            |          HUB address           |
'' -------------------------------------------------------------------
''
'' Read/Write to the RAM currently is a bit of a bodge.  Mainly it is just the code from
'' the original Dracblade.spin patched in.  The original text strings "R", "W" etc are
'' translated in the spin layer at the moment, ideally the calls should be altered to use
'' the codes understood by PASM.  The response on success sets cmd to 0 but doesn't update
'' the address/counter fields as the SPI driver does.


'' SPI data is handled a little differently.  For ioSpiInit, the 6 bit pin numbers for DO,
'' Clk, DI, and CS are given from MSB to LSB of the 24 bit address field of the command and
'' are used for all further I/O operations (until an ioSpiStop is done).

CON
'' Command code and error information for I2C driver
'' (For convenience in using just OS_loaderInit, these are included here.  The "master"
''  copies are considered the ones in OS_loader and these must be kept up-to-date).

  ioRamRead   = %00000001                      ' Read from EEPROM to HUB RAM (16 bit addresses)
  ioRamWrite  = %00000010                      ' Write to EEPROM from HUB RAM (16 bit addresses)
  ioLedOn  = %00000011                      ' Read from a device with only 8-bit addresses
  ioLedOff = %00000100                      ' Write to a device with only 8-bit addresses
  ioRamHigh  = %00000101                      ' Read from a device without address bytes
'  ioWrite0Cmd = %00000110                      ' Write to a device without address bytes
'  ioBootCmd   = %00001000                      ' Read from EEPROM to HUB RAM, then start a
                                               '  new SPIN interpreter in the COG whose ID is
                                               '  supplied in the lower 3 bits of this command
                                               '  This COG is stopped before the read is done
                                               '  unless it's the one used to execute the loader
  ioSpiInit   = %00010000                      ' Initialize the specified SPI bus and SD card
  ioSpiStop   = %00010001                      ' Change all SD card pins to inputs
  ioSpiRead   = %00010010                      ' Read one or more bytes from the SD card
  ioSpiWrite  = %00010011                      ' Write one or more bytes from the SD card
  ioSpiBoot   = %00011000                      ' Like ioBootCmd, but uses ioSpiRead for loading
  ioCmdMask   = %00011111                      ' Used to mask off command bits
  ioSpiMask   = %00010000                      ' Used to test for SPI command codes

' Options for commands

  ioNoStore   = %00100000                      ' If set, data is not stored into main memory
                                               '  If ioBootCmd or ioSpiBoot, no cogs are
                                               '   stopped and a new cog is not started.
  ioLowSpeed  = %01000000                      ' If set, I2C runs at 100KHz rather than 400KHz
  ioStopLdr   = %10000000                      ' If set, the loader's cog is stopped after a boot

' Return status

  ioWriteErr  = %10000000                      ' An error occurred during an I2C write (NAK)

  ioTestRdy   = ioCmdMask << 24                ' Used to test 1st control long for ready
  ioTestErr   = ioWriteErr << 24               ' Used to test 1st control long for write error

'' Other constants from OS_loader

  i2cBootSCL  = 28                             ' Boot EEPROM SCL pin
  bootAddr    = i2cBootSCL << 18               ' Address of boot EEPROM
  clkfreqVal  = $0000                          ' Current CLKFREQ value stored here
  clksetVal   = $0004                          ' Current CLKSET value stored here
  chksumVal   = $0005                          ' Checksum over memory stored here
  vbase       = $0008                          ' Length of Spin program loaded (# longs * 4)
  dbase       = $000A                          ' Address of start of stack (marker below)

VAR
  long cog, control

PUB initSDCard(DO,Clk,DI,CS) | t                       '' Initialize SD card access
  t := cnt
  repeat while long[control][0] & ioTestRdy            ' Wait for previous I/O to finish
  long[control][1] := 0
  long[control][0] := ioSpiInit << 24 | DO << 18 | Clk << 12 | DI << 6 | CS
  repeat while long[control][0] & ioTestRdy            ' Wait for this to finish
  return (long[control][0] & ioTestErr) <> 0           ' Return any error code

PUB stopSDCard                                         '' Stop SD card access
  repeat while long[control][0] & ioTestRdy            ' Wait for previous I/O to finish
  long[control][1] := 0
  long[control][0] := ioSpiStop << 24
  repeat while long[control][0] & ioTestRdy            ' Wait for this to finish
  return (long[control][0] & ioTestErr) <> 0           ' Return any error code

PUB readSDCard(addr,buffer,count)                      '' Read block(s) from SD card to RAM
  repeat while long[control][0] & ioTestRdy            ' Wait for previous I/O to finish
  long[control][1] := (count << 16) | (buffer & $FFFF)
  long[control][0] := (ioSpiRead << 24) | (addr & $FFFFFF)
  repeat while long[control][0] & ioTestRdy            ' Wait for this to finish
  return (long[control][0] & ioTestErr) <> 0           ' Return any error code

PUB writeSDCard(addr,buffer,count)                     '' Write block(s) to SD card from RAM
  repeat while long[control][0] & ioTestRdy            ' Wait for previous I/O to finish
  long[control][1] := (count << 16) | (buffer & $FFFF)
  long[control][0] := (ioSpiWrite << 24) | (addr & $FFFFFF)
  repeat while long[control][0] & ioTestRdy            ' Wait for this to finish
  return (long[control][0] & ioTestErr) <> 0           ' Return any error code

PUB start(ctrlBlk) | t                                 '' Start the I2C I/O driver (standalone)
  control := ctrlBlk                                   ''  using address of 2 longs for control
  stop                                                 ' Stop a previous copy
  'computeTimes
  i2cPause    := clkfreq / 100000                      ' Pause between checks for operations
  long[control][0] := 0                                ' Initialize the control block
  long[control][1] := 0
  cog := cognew(@tbp2_start,control) + 1            ' Start a new cog with the I2C driver
  return cog > 0                                       ' Indicate success

PUB stop                                               '' Stop the I2C I/O driver (standalone)
  if cog > 0
    cogstop(cog - 1)

PUB getControl(i)                                      '' Return a long from the control block
  return long[control][i]                              ' Check for operation completed first

PUB setControl(i,value)                                '' Set value of a long in the control block
  long[control][i] := value                            ' Always set the first long last
PUB location
   return @tbp2_start  ' returns the location of the hub space used by the sdspi port driver

PUB DoCmd(command_, hub_address, ram_address, block_length) : err_ | command2_
' Do the command: R, W, N, F, H
  if command_ == "R"
    command2_ := ioRamRead
  if command_ == "W"
    command2_ := ioRamWrite
  if command_ == "N"
    command2_ := ioLedOn
  if command_ == "F"
    command2_ := ioLedOff
  if command_ == "H"
    command2_ := ioRamHigh

  repeat while long[control][0] & ioTestRdy
  long[control][1] := ((block_length & $ffff) << 16) + (hub_address & $ffff)
  long[control][0] := ((command2_ & $ff) << 24) + (ram_address & $ffffff)
  'hubaddrs := hub_address       ' hub address start
  'ramaddrs := ram_address       ' ram address start
  'blocklen := block_length      ' block length
  'command  := command_          ' must be last !!
' Wait for command to complete and get status
  repeat while long[control][0] & ioTestRdy          ' driver cog sets =0 when done
  err_ := long[control][0]                  ' driver cog sets =0 if no error, else xx = error code

DAT
                        org     0

tbp2_start    ' setup the pointers to the hub command interface (saves execution time later
err                     nop                     ' -+  error status returned (=0=false=good)


' Initialise hardware (unlike the triblade, just tristates everything and read/write set the pins)
init                    mov     err, #0                  ' reset err=false=good
                        andn     dira,LatchDirection                ' tristate the pins

done                    wrlong  err, PAR             ' status  =0=false=good, else error x
                        mov     i2cTemp,PAR
                        add     i2cTemp, #4
                        wrlong  zero, i2cTemp            ' command =0 (done)
' wait for a command (pause short time to reduce power)
pause                   mov     ctr, delay      wz      ' if =0 no pause
              if_nz     add     ctr, cnt
              if_nz     waitcnt ctr, #0                 ' wait for a short time (reduces power)
'                        rdlong  cmd, comptr     wz      ' command ?
'              if_z      jmp     #pause                  ' not yet
' decode command



i2cNewOpFetch           rdlong  i2cAddr,PAR             ' Fetch control information
                        mov     i2cCmd,i2cAddr
                        shr     i2cCmd,#24              ' Isolate command code
                        mov     Options,i2cCmd
                        and     i2cAddr,i2cAddrMask     ' Only need address at this point
                        and     i2cCmd,#ioCmdMask  wz
                if_z    jmp     #pause          ' Wait for a new operation

                        mov     i2cTemp,PAR
                        add     i2cTemp,#4              ' Now get 2nd long of packet
                        rdlong  i2cCount,i2cTemp
                        mov     i2cBufAdr,i2cCount      ' Byte count
                        rdlong  SaveClkFreq,#clkfreqVal ' Save clock frequency and mode
                        shr     i2cCount,#16
                        and     i2cBufAdr,i2cWordMask   ' HUB RAM address of buffer
                        rdbyte  SaveClkMode,#clksetVal
                        movs    ShiftData,#0            ' Initialize for saving Preamble
                        test    i2cCmd, #ioSpiMask      wz
                if_nz   jmp     #spiEntryPoint          ' if it's not an SPI command do RAM
                        'jmp     #spiEntryPoint

ramEntryPoint
                        cmp     i2cCmd, #ioRamRead       wz      ' R = read block
              if_z      jmp     #rdblock
                        cmp     i2cCmd, #ioRamWrite       wz      ' W = write block
              if_z      jmp     #wrblock
                        cmp     i2cCmd, #ioLedOn       wz      ' N= led on
              if_z      jmp     #led_turn_on
                        cmp     i2cCmd, #ioLedOff       wz      ' F = led off
              if_z      jmp     #led_turn_off
                        cmp     i2cCmd, #ioRamHigh       wz      ' H sets the high latch
              if_z      jmp     #sethighlatch
                        mov     err, i2cCmd                ' error = cmd (unknown command)
                        jmp     #done

'' SPI routines for Rokicki's SD card FAT file system driver

spiEntryPoint           'test    i2cCmd,#ioBootCmd  wc   ' Check for boot
              'if_c      jmp     #spiDoRead              ' (Treat like read)
                        cmp     i2cCmd,#ioSpiStop wc,wz
              if_c      jmp     #spiDoInit              ' Decode operation
              if_z      jmp     #spiDoStop
                        cmp     i2cCmd,#ioSpiWrite wc
              if_c      jmp     #spiDoRead
                        jmp     #spiDoWrite

'' Initialize SPI communications.  The pin numbers of the 4 I/O pins are
'' provided in the 24 bit address field of the control packet.  From MSB to
'' LSB, these are DO - Data Out, Clk - Clock, DI - Data In, CS - Card Select.

spiDoInit               movd    :moveIt,#spiMaskCS
                        mov     spiBlkCnt,#4
:makeMask               mov     i2cMask,#1
                        mov     i2cTemp,i2cAddr         ' Only use lower 5 bits of
                        and     i2cTemp,#%11111         '  6 bit shift count field
                        shl     i2cMask,i2cTemp
:moveIt                 mov     0-0,i2cMask             ' Store the bit mask for the pin
                        cmp     spiBlkCnt,#1       wz
              if_ne     or      outa,i2cMask            ' Make all pins high outputs
              if_ne     or      dira,i2cMask            '  except DO is an input since
              if_e      andn    dira,i2cMask            '   input/output is card relative
                        sub     :moveIt,incrDst
                        ror     i2cAddr,#6
                        djnz    spiBlkCnt,#:makeMask
                        rol     i2cAddr,#24             ' Leave i2cAddr unchanged
                        mov     i2cTime,cnt             ' Set up a 1 second timeout
                        mov     spiBlkCnt,spiInitCnt
:initRead               call    #spiRecvByte            ' Output a stream of 32K clocks
                        djnz    spiBlkCnt,#:initRead    '  in case SD card left in some
                        mov     spiOp,#0                '   undefined state
                        mov     spiParm,#0
                        call    #spiSendCmd             ' Send a reset command and deselect
                        or      outa,spiMaskCS          '  to get SD card into SPI mode
:waitIdle               mov     spiOp,#55
                        call    #spiSendCmd             ' APP_CMD (Application Specific)
                        mov     spiOp,#41
                        call    #spiSendCmd             ' SEND_OP_COND (Initialization)
                        or      outa,spiMaskCS
                        cmp     i2cData,#1         wz   ' Wait until response not In Idle
              if_e      jmp     #:waitIdle
                        tjz     i2cData,#i2cGoUpdate    ' Initialization complete
                        or      i2cAddr,errorFlag
                        jmp     #i2cGoUpdate            ' Could not initialize the card

'' Stop SPI communications.  Any previously used I/O pins are set to input
'' mode and the masks for the I/O pins are zeroed.

spiDoStop
                        andn    dira,spiMaskCS          ' This is used for the situation
                        mov     spiMaskCS,#0            '  where the pins used for the SD
                        andn    dira,spiMaskDI          '   card may be used for some other
                        mov     spiMaskDI,#0            '    purpose when the card is removed
                        andn    dira,spiMaskClk
                        mov     spiMaskClk,#0
                        andn    dira,spiMaskDO
                        mov     spiMaskDO,#0
                        jmp     #i2cGoUpdate


{{
' Modification by Cluso99 to force the SD card to tristate the D0 pin (allows other chips on the bus)
spiDoStop               or      outa,spiMaskCS          ' CS=1
                        andn    dira,spiMaskDO          ' disable D0 as output (s/be input anyway)
                        call    #spiRecvByte            ' Output a stream of clocks 
                        andn    dira,spiMaskCS          ' CS input
                        mov     spiMaskCS,#0            ' 
                        mov     spiMaskDO,#0            ' D0 already input (above)
                        andn    dira,spiMaskDI          ' DI input
                        mov     spiMaskDI,#0            '  
                        andn    dira,spiMaskClk         ' CLK input
                        mov     spiMaskClk,#0
                        jmp     #i2cGoUpdate

}}
'' Read one or more 512 byte blocks and store the specified number of bytes
'' into the HUB location given.  The block number is provided in the 24 bit
'' address field and incremented after every block is read.  Partial blocks are
'' allowed and any extra bytes read are discarded.

spiDoRead               mov     spiOp,#17               ' READ_SINGLE_BLOCK
:readRepeat             mov     i2cTime,cnt             ' Save start of timeout
                        mov     spiParm,i2cAddr
                        call    #spiSendCmd             ' Read from specified block
                        call    #spiResponse
                        mov     spiBlkCnt,spiBlkSize    ' Transfer a block at a time
:getRead                call    #spiRecvByte
                        tjz     i2cCount,#:skipStore    ' Check for count exhausted
                        call    #StoreData
                        sub     i2cCount,#1
:skipStore              djnz    spiBlkCnt,#:getRead     ' Are we done with the block?
                        call    #spiRecvByte
                        call    #spiRecvByte            ' Yes, finish with 16 clocks
                        add     i2cAddr,#1
                        or      outa,spiMaskCS          ' Increment address, deselect card
                        tjnz    i2cCount,#:readRepeat   '  and check for more blocks to do
checkEndIO              {test    i2cCmd,#ioBootCmd  wc
                 if_nc  }jmp     #i2cGoUpdate            ' If not booting, we're done
                        {test    i2cAddr,errorFlag  wc
                        and     CheckSum,#$FF      wz   ' If booting, no errors can occur
           if_z_and_nc  jmp     #nowBootSpin            '  and checksum must be zero
                        or      i2cAddr,errorFlag
                        test    Options,#noStore   wc
                 if_c   jmp     #i2cGoUpdate            ' Return error status if noStore
                         }
'stopThisCOG             cogid   i2cCogId                ' If an unrecoverable error occurs,
'                        cogstop i2cCogId                '  stop this cog
                      
'' Write one or more 512 byte blocks with the specified number of bytes from
'' the HUB location given.  The block number is provided in the 24 bit address
'' field and incremented after every block is written.  Partial blocks are
'' allowed and are padded with zeroes.

spiDoWrite              mov     spiOp,#24               ' WRITE_BLOCK
                        mov     i2cTime,cnt             ' Setup timeout
                        mov     spiParm,i2cAddr
                        call    #spiSendCmd             ' Write to specified block
                        mov     i2cData,#$FE            ' Ask to start data transfer
                        call    #spiSendByte
                        mov     spiBlkCnt,spiBlkSize    ' Transfer a block at a time
:putWrite               mov     i2cData,#0              '  padding with zeroes if needed
                        tjz     i2cCount,#:padWrite     ' Check for count exhausted
                        rdbyte  i2cData,i2cBufAdr       ' If not, get the next data byte
                        add     i2cBufAdr,#1
                        sub     i2cCount,#1
:padWrite               call    #spiSendByte
                        djnz    spiBlkCnt,#:putWrite    ' Are we done with the block?
                        call    #spiRecvByte
                        call    #spiRecvByte            ' Yes, finish with 16 clocks
                        call    #spiResponse
                        and     i2cData,#$1F            ' Check the response status
                        cmp     i2cData,#5         wz
              if_ne     or      i2cAddr,errorFlag       ' Must be Data Accepted
              if_ne     jmp     #i2cGoUpdate
                        movs    spiWaitData,#0          ' Wait until not busy
                        call    #spiWaitBusy
                        add     i2cAddr,#1
                        or      outa,spiMaskCS          ' Increment block address and go
                        tjnz    i2cCount,#spiDoWrite    '  to next if more data remains
                        jmp     #i2cGoUpdate

'' Mid level SPI I/O

spiSendCmd              andn    outa,spiMaskCS          ' Send command sequence.  Begin by
                        call    #spiRecvByte            '  selecting card and clocking
                        mov     i2cData,spiOp
                        or      i2cData,#$40            ' Send command byte (1st 2 bits %01)
                        call    #spiSendByte
                        mov     i2cData,spiParm
                        shr     i2cData,#15             ' Supplied address is sector number
                        call    #spiSendByte 
                        mov     i2cData,spiParm         ' Send to SD card as byte address,
                        shr     i2cData,#7              '  in multiples of 512 bytes
                        call    #spiSendByte
                        mov     i2cData,spiParm         ' Total length of this address is
                        shl     i2cData,#1              '  four bytes
                        call    #spiSendByte
                        mov     i2cData,#0
                        call    #spiSendByte
                        mov     i2cData,#$95            ' CRC code (for 1st command only)
                        call    #spiSendByte
spiResponse             movs    spiWaitData,#$FF        ' Wait for response from card
spiWaitBusy             call    #spiRecvByte
                        mov     i2cTemp,cnt
                        sub     i2cTemp,i2cTime         ' Check for expired timeout (1 sec)
                        cmp     i2cTemp,SaveClkFreq wc
              if_nc     or      i2cAddr,errorFlag
              if_nc     jmp     #i2cGoUpdate
spiWaitData             cmp     i2cData,#0-0       wz   ' Wait for some other response
              if_e      jmp     #spiWaitBusy            '  than that specified
spiSendCmd_ret
spiResponse_ret
spiWaitBusy_ret         ret

'' Low level byte I/O

spiSendByte             mov     i2cMask,#%10000000
:sendBit                test    i2cMask,i2cData    wc
                        andn    outa,spiMaskClk         ' Send data bytes MSB first
                        muxc    outa,spiMaskDI
                        or      outa,spiMaskClk
                        shr     i2cMask,#1              ' When mask shifted out, we're done
                        tjnz    i2cMask,#:sendBit
                        or      outa,spiMaskDI          ' Leave DI in idle (high) state
spiSendByte_ret         ret

spiRecvByte             mov     i2cMask,#%10000000
:recvBit                andn    outa,spiMaskClk         ' Receive data bytes MSB first
                        or      outa,spiMaskClk         ' Copy DO to data bit
                        test    spiMaskDO,ina      wc     
                        muxc    i2cData,i2cMask
                        shr     i2cMask,#1              ' When mask shifted out, we're done
                        tjnz    i2cMask,#:recvBit
                        and     i2cData,#%11111111      ' Eight bits received
spiRecvByte_ret         ret

'' For both I2C and SPI, store data on a read operation unless ioNoStore is set.
'' Accumulate a checksum and always save a copy of the first 16 bytes read.
'' If this is an ioBootCmd or ioSpiBoot, adjust the amount to be read based
'' on the value in the program preamble in the word at vbase ($0008).

StoreData               'test    Options,#ioNoStore wc
                {if_nc}   wrbyte  i2cData,i2cBufAdr       ' Store data in specified location
                        add     i2cBufAdr,#1            '  and increment the address
'                        add     CheckSum,i2cData        ' Accumulate checksum for ioBootCmd
ShiftData               shl     i2cData,#0-0
StoreLocal              'or      Preamble+0,i2cData      ' Store a local copy of the program
                        add     ShiftData,#8            '  preamble for when we're reading
                        cmp     ShiftData,testIns  wz   '   in a new Spin program
                if_z    movs    ShiftData,#0            ' Pack the data into successive longs
StoreData_ret           ret

i2cGoUpdate             and     i2cBufAdr,i2cWordMask   ' Copy updated information
                        shl     i2cCount,#16            '  back to control packet
                        or      i2cCount,i2cBufAdr
                        mov     i2cTemp,PAR
                        add     i2cTemp,#4
                        wrlong  i2cCount,i2cTemp
                        wrlong  i2cAddr,PAR             ' Indicate operation is done
                        jmp     #pause          '  and go wait for a new one

'=========================================================================================================
' RAM functions
'=========================================================================================================

' turn led on
led_turn_on             or      HighLatch,ledpin        ' set the led pin high
                        jmp     #OutputHighLatch         ' send this out

led_turn_off            andn    HighLatch,ledpin        ' set the led pin low
                        jmp     #OutputHighLatch         ' send this out

' set high address bytes with command H, pass value in third variable of the DoCmd
' 4 bytes - masks off all but bits 16 to 23

sethighlatch            'call #ram_open                  ' gets address value in 'address'
                        shr  i2cAddr,#16                ' shift right by 16 places
                        and  i2cAddr,#$FF               ' ensure rest of bits zero
                        mov  HighLatch,i2cAddr          ' put value into HighLatch
                        jmp  #OutputHighLatch           ' and output it

'---------------------------------------------------------------------------------------------------------
'Memory Access Functions

rdblock                 'call    #ram_open               ' get variables from hub variables
rdloop                  'call    #read_memory_byte       ' read byte from address into data_8
' put this code inline to remove the call since this is the only place it gets called from
read_memory_byte        call #RamAddress                ' sets up the latches with the correct ram address
                        ' default behaviour should have dira already set up correctly now
                        'or dira,LatchDirection        ' for reads so P0-P7 tristate till do read
                        ' need to turn bus around after writing address
                        andn dira, #$ff
                        andn outa, LatchDirection       ' clear address bits (read = 0000 on 138)
                        '' !!! next command seems redundant but not sure
                        'or outa,GateHigh                ' actually ReadEnable but they are the same
                        andn outa,GateHigh              ' set gate low
                        nop                             ' short delay to stabilise
                        nop
                        mov data_8, ina                 ' read SRAM
                        and data_8, #$FF                ' extract 8 bits
                        or  outa,GateHigh               ' set the gate high again
'read_memory_byte_ret    ret
                        wrbyte  data_8,i2cBufAdr          ' write data_8 to hubaddr ie copy byte to hub
                        add     i2cBufAdr,#1              ' add 1 to hub address
                        add     i2cAddr,#1              ' add 1 to ram address
                        djnz    i2cCount,#rdloop             ' loop until done
                        jmp     #init                   ' set error=False

wrblock                 'call    #ram_open
wrloop                  rdbyte  data_8, i2cBufAdr         ' copy byte from hub
                        'call    #write_memory_byte      ' write byte from data_8 to address
' moved nline as this was only called from here
write_memory_byte       call #RamAddress                ' sets up the latches with the correct ram address
                        mov outx,data_8                 ' get the byte to output
                        and outx, #$FF                  ' ensure upper bytes=0
                        or outx,WriteEnable             ' or with correct 138 address
                        andn outa,latchdirection
                        or outa,outx                   ' send it out
                        andn outa,GateHigh              ' set gate low
                        nop                             ' no nop doesn't work, one does, so put in two to be sure
                        nop                             ' another NOP
                        or outa,GateHigh                ' set it high again
'write_memory_byte_ret   ret
                        add     i2cBufAdr,#1              ' add 1 to hub address
                        add     i2cAddr,#1              ' add 1 to ram address
                        djnz    i2cCount,#wrloop             ' loop until done
                        jmp     #init                   ' reinitialise

RamAddress ' sets up the ram latches. Assumes high latch A16-A18 low so only accesses 64k of ram
                        or dira,LatchDirection         ' set up the pins for programming latch chips
                        mov outx,i2cAddr                ' get the address into a temp variable
                        and outx,#$FF                   ' mask the low byte
                        or  outx,LowAddress             ' or with 138 low address
                        andn outa, LatchDirection
                        or outa,outx                   ' send it out
                        andn outa,GateHigh              ' set gate low
                                                        ' ?? a NOP
                        or outa,GateHigh                ' set it high again
                                                        ' now repeat for the middle byte
                        mov outx,i2cAddr                ' get the address into a temp variable
                        shr outx,#8                     ' shift right by 8 places
                        and outx,#$FF                   ' mask the low byte
                        or  outx,MiddleAddress          ' or with 138 middle address
                        andn outa, LatchDirection
                        or outa,outx                   ' send it out
                        andn outa,GateHigh              ' set gate low
                        or outa,GateHigh                ' set it high again
RamAddress_ret          ret

OutputHighLatch ' sends out HighLatch to the 374 that does A16-19, led and the 4 spare outputs
                        or      dira,latchdirection     ' setup active pins 138 and bus
                        andn    outa, latchdirection
                        or      outa,HighLatch          ' send out HighLatch
                        or      outa,HighAddress        ' or with the high address
                        andn    outa,GateHigh           ' set gate low
                        or      outa,GateHigh           ' set the gate high again
OutputHighLatch_ret     jmp     #Init                   ' retrun error=False





delay                   long    80                                    ' waitcnt delay to reduce power (#80 = 1uS approx)
ctr                     long    0                                     ' used to pause execution (lower power use) & byte counter
GateHigh                long    %00000000_00000000_00000001_00000000  ' HC138 gate high, all others must be low
Outx                    long    0                                     ' for temp use, same as n in the spin code
LatchDirection          long    %00000000_00000000_00001111_11111111 ' 138 active, gate active and 8 data lines active
LatchDirection2         long    %00000000_00000000_00001111_00000000 ' for reads so data lines are tristate till the read
LowAddress              long    %00000000_00000000_00000101_00000000 ' low address latch = xxxx010x and gate high xxxxxxx1
MiddleAddress           long    %00000000_00000000_00000111_00000000 ' middle address latch = xxxx011x and gate high xxxxxxx1
HighAddress             long    %00000000_00000000_00001001_00000000 ' high address latch = xxxx100x and gate high xxxxxxx1
'ReadEnable long    %00000000_00000000_00000001_00000000 ' /RD = xxxx000x and gate high xxxxxxx1
                                                        ' commented out as the same as GateHigh
WriteEnable             long    %00000000_00000000_00000011_00000000 ' /WE = xxxx001x and gate high xxxxxxx1
Zero                    long    %00000000_00000000_00000000_00000000 ' for tristating all pins
data_8                  long    %00000000_00000000_00000000_00000000 ' so code compatability with zicog driver
'address                 long    %00000000_00000000_00000000_00000000 ' address for ram chip
ledpin                  long    %00000000_00000000_00000000_00001000 ' to turn on led
HighLatch               long    %00000000_00000000_00000000_00000000 ' static value for the 374 latch that does the led, hA16-A19 and the other 4 outputs
'' This action table contains bit sequences for controlling device addressing and read/write
'' mode selection for each of the commands possible.  From LSB to MSB, the actions are:
'' 0 - Write the EEPROM device select code for write mode and 2 address bytes
'' 1 - Write the MSB device address or (for ioRead1Cmd/ioWrite1Cmd) a device select code
'' 2 - Write the LSB device address or (for ioRead0Cmd/ioWrite0Cmd) a device select code
'' 3 - Output a Start Sequence prior to reselecting in read mode
'' 4 - Fetch a data value for writing
'' 5 - Construct an EEPROM device select code for read mode and 2 address bytes
'' 6 - Construct a read mode device select code from the MSB of the 16 bit device address
'' 7 - Write the data value or read mode device select code
'' 8 - Read a byte of data from the device and store it

i2cZero
ActionTbl               long    %0000000000             ' Command not used (indicates done)
                        long    %0110101111,%0010010111 ' Read/Write with 2 bytes of addressing
                        long    %0111001110,%0010010110 ' Read/Write with 1 byte of addressing
                        long    %0100000100,%0010010100 ' Read/Write data only

'' Constants for all routines

i2cWordMask             long    $0000FFFF
i2cAddrMask             long    $00FFFFFF
errorFlag               long    $80000000               ' NAK received during write cycle
speedMask               long    $40000000               ' One if 100KHz bus, zero if 400KHz
time_xtal               long    20 * 20000 / 4 / 1      ' 20ms (@20MHz, 1 inst/loop)
'interpreter             long    ($0004 << 16) | ($F004 << 2) | %0000
i2cBootSCLm             long    |<i2cBootSCL            ' Bit mask for pin 28 SCL use
spiBlkSize                                              ' Number of bytes in an SD card block
incrDst                 long    %10_00000000            ' Used to increment destination field
testIns                 shl     i2cData,#32             ' Used to compare for end of word packing
'initStore               or      Preamble+0,i2cData      ' Used to initialize packing instruction
'testDst                 or      Preamble+4,i2cData      ' Used to check for end of packing buffer
noStore                 jmp     #StoreData_ret          ' Used after all data stored into Preamble
spiInitCnt              long    32768 / 8               ' Initial SPI clocks produced
StackMark               long    $FFF9FFFF               ' Two of these mark the base of the stack

'' Variables for all routines

'Preamble                long    0, 0, 0, 0              ' Private copy of program preamble
Action                  long    0
'i2cOther
'i2cCogId                long    0
i2cCmd                  long    0
'FirstCall               long    -1                      ' One if I2C pins not initialized yet
i2cTemp                 long    0
i2cCount                long    0
i2cBufAdr               long    0
i2cAddr                 long    0
'i2cDataSet              long    0                       ' Minumum data setup time (ticks)
'i2cClkLow               long    0                       ' Minimum clock low time (ticks)
'i2cClkHigh              long    0                       ' Minimum clock high time (ticks)
'i2cDataSet1             long    0                       ' Minumum data setup time (ticks) 100KHz
'i2cClkLow1              long    0                       ' Minimum clock low time (ticks) 100KHz
'i2cClkHigh1             long    0                       ' Minimum clock high time (ticks) 100KHz
'i2cDataSet4             long    0                       ' Minumum data setup time (ticks) 400KHz
'i2cClkLow4              long    0                       ' Minimum clock low time (ticks) 400KHz
'i2cClkHigh4             long    0                       ' Minimum clock high time (ticks) 400KHz
i2cPause                long    0                       ' Pause before re-fetching next operation
SaveClkFreq             long    0                       ' Initial clock frequency (clkfreqVal)
SaveClkMode             long    0                       ' Initial clock mode value (clksetVal)
spiBlkCnt               long    0                       ' Number of SD card bytes to go in block
'CheckSum                long    0                       ' Checksum of bytes for ioBootCmd
Options                 long    0                       ' Option bits (ioNoStore, ioLowSpeed)

'' Local variables for low level I2C routines

spiOp                                                   ' Operation code for SPI command
i2cSCL                  long    0                       ' Bit mask for SCL
spiParm                                                 ' Parameter value for SPI command
i2cSDA                  long    0                       ' Bit mask for SDA
i2cTime                 long    0                       ' Used for timekeeping
i2cData                 long    0                       ' Data to be transmitted / received
i2cMask                 long    0                       ' Bit mask for bit to be tx / rx
i2cBitCnt               long    0                       ' Number of bits to tx / rx

'' Additional local variables for SPI SD Card access

spiMaskDO               long    0
spiMaskClk              long    0
spiMaskDI               long    0
spiMaskCS               long    0

                        fit     $170
